"""
Copyright 2019 de la Dirección General de Sistemas Informáticos – Secretaría Legal y Técnica - Nación.

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see http://www.gnu.org/licenses/
"""
from datetime import datetime
import logging
import zlib

from gateway.gateway import ContractInteractor, Gateway
import json
import hashlib

from gateway.utils import hex32bytes_string

logger = logging.getLogger('logger')


def resolve_block_number(block_number):
    try:
        block = int(block_number.hex(), 16)
    except AttributeError as ae:
        block = block_number
    return block


class CompressSerializer:
    def serialize(self, data):
        return zlib.compress(json.dumps(data).encode('utf-8'))

    def deserialize(self, serialized):
        return json.loads(zlib.decompress(serialized).decode())


class DataSerializer:
    def serialize(self, data):
        return hashlib.sha256(json.dumps(data).encode('utf-8')).digest()


class InvalidTrackingIdException(Exception):
    pass


class TrackStarter:
    contract_interactor = ...  # type: ContractInteractor
    serializer = DataSerializer()

    def __init__(self, contract_interactor, from_account):
        self.contract_interactor = contract_interactor
        self.owner = from_account

    def start(self, tracking_id, state):
        """

        :type tracking_id: bytes
        """
        if len(tracking_id) > 32:
            raise InvalidTrackingIdException("Tracking id is {} bytes long, more than 32".format(str(len(tracking_id))))
        estado_serialized = self.serializer.serialize(state)
        function = self.contract_interactor.get_function('startTrackWorkflow')
        logger.info("Starting track id: {} / state : {} / serialized state {}"
                    .format(str(tracking_id.hex()), str(state), str(estado_serialized)))
        res = self.contract_interactor.execute_function(function(tracking_id, estado_serialized),
                                                        {'from': self.owner, 'gas': 2000000})
        return res


class ParticipantsLister:
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, contract_interactor, from_block=0, to_block='latest'):
        self.to_block = to_block
        self.from_block = from_block
        self.contract_interactor = contract_interactor

    def all(self):
        logger.info("Listing all participants")
        entries = self.contract_interactor.search_logs('ParticipantAdded',
                                                       {'fromBlock': self.from_block, 'toBlock': self.to_block})
        dict_res = {}
        res = []
        for e in entries:
            dict_res.setdefault(e['args']['participant'], [])
            dict_res[e['args']['participant']].append(e['args']['role'].decode('utf-8').strip('\\\u0000'))
        for k, v in dict_res.items():
            res.append({'participant': k, 'roles': list(set(v))})  # para sacar los repetidos
        return res

    def is_participant(self, account):
        function = self.contract_interactor.get_function('isParticipant')
        return self.contract_interactor.call_function(function(account), {})


class RolesLister:
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, contract_interactor, from_block=0, to_block='latest'):
        self.to_block = to_block
        self.from_block = from_block
        self.contract_interactor = contract_interactor

    def all(self):
        logger.info("Listing all roles")
        entries = self.contract_interactor.search_logs('ParticipantAdded',
                                                       {'fromBlock': self.from_block, 'toBlock': self.to_block})
        dict_res = {}
        for e in entries:
            dict_res.setdefault(e['args']['role'].decode('utf-8').strip('\\\u0000'), [])
            dict_res[e['args']['role'].decode('utf-8').strip('\\\u0000')].append(e['args']['participant'])
        return dict_res


class ParticipantAdding:
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, contract_interactor, owner):
        self.contract_interactor = contract_interactor
        self.owner = owner

    def add_participant(self, address, role):
        function = self.contract_interactor.get_function('addParticipantRole')
        logger.info("Adding participant address: {} / role : {}".format(str(address), str(role)))
        res = self.contract_interactor.execute_function(function(role.encode('utf-8'), address),
                                                        {'from': self.owner})
        return res


class WorkflowStepAdding:

    def __init__(self, contract_interactor, owner):
        self.contract_interactor = contract_interactor
        self.owner = owner

    def add_workflow_step(self, message, role):
        function = self.contract_interactor.get_function('addWorkflowStep')
        logger.info("Adding workflow step message: {} / role : {}".format(str(message), str(role)))
        res = self.contract_interactor.execute_function(function(message, role.encode('utf-8')),
                                                        {'from': self.owner})
        return res


class WorkflowLister:
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, contract_interactor, from_block=0, to_block='latest'):
        self.to_block = to_block
        self.from_block = from_block
        self.contract_interactor = contract_interactor

    def workflow_is_defined(self):
        logger.info("Checking if workflow is defined")
        function = self.contract_interactor.get_function('workflowIsDefined')
        res = self.contract_interactor.call_function(function())
        logger.debug("Workflow is defined? {}".format(str(res)))
        return res

    def all(self):
        logger.info("Listing all workflow steps")
        entries = self.contract_interactor.search_logs('WorkflowStepAdded',
                                                       {'fromBlock': self.from_block, 'toBlock': self.to_block})
        res = []
        for e in entries:
            res.append({'role': e['args']['role'].decode('utf-8').strip('\\\u0000'), 'message': e['args']['message']})
        return res


class TrackingAssociation:
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, contract_interactor, from_block=0, to_block='latest'):
        self.from_block = from_block
        self.to_block = to_block
        self.contract_interactor = contract_interactor

    def tracking_contract(self, tracking_id):
        logger.info("Searching tracking address for tracking id {}".format(str(tracking_id.hex())))
        # tracking_id = string_to_hex32bytes(tracking_id)
        entries = self.contract_interactor.search_logs('TrackingStarted',
                                                       {'fromBlock': self.from_block, 'toBlock': self.to_block,
                                                        'argument_filters': {
                                                            'trackingId': tracking_id
                                                        }
                                                        })
        if len(entries) == 0:
            raise NoTrackingException("No tracking for this id")

        return entries[0]['args']['trackingAddress']


class TrackForwarder:
    contract_interactor = ...  # type: ContractInteractor
    serializer = DataSerializer()

    def __init__(self, contract_interactor, owner, tracking_associator):
        self.tracking_associator = tracking_associator
        self.contract_interactor = contract_interactor
        self.owner = owner

    def move_forward(self, tracking_id, estado=None):
        if estado is None:
            estado = b''
        else:
            estado = self.serializer.serialize(estado)
        function = self.contract_interactor.get_function('forwardWorkflow')
        tracking_address = self.tracking_associator.tracking_contract(tracking_id)
        logger.info("Moving forward tracking id : {} / state : {}".format(str(tracking_id.hex()), str(estado)))
        res = self.contract_interactor.execute_function(function(tracking_address, estado), {'from': self.owner})
        return res


class NoTrackingException(Exception):
    pass


class TrackStatus:
    contract_interactor = ...  # type: ContractInteractor
    tracking_association = ...  # type: TrackingAssociation
    gateway = ...  # type: Gateway
    serializer = DataSerializer()

    def __init__(self, contract_interactor, tracking_association, from_block=0, to_block='latest'):
        self.tracking_association = tracking_association
        self.from_block = from_block
        self.to_block = to_block
        self.contract_interactor = contract_interactor
        self.gateway = self.contract_interactor.gateway

    def last_message(self, tracking_id):
        logger.info("Searching for last message tracking id : {}".format(str(tracking_id.hex())))
        tracking_history = self.tracking_history(tracking_id)
        if len(tracking_history) == 0:
            raise NoTrackingException("No tracking existed with that id")
        return tracking_history[-1]['message']

    def is_this_state(self, tracking_id, state):
        hashstate_function = self.contract_interactor.get_function('currentHashState')
        try:
            tracking_address = self.tracking_association.tracking_contract(tracking_id)
            logger.info("Checking tracking id {} has state {}".format(str(tracking_id), str(state)))
            hashstate = self.contract_interactor.call_function(hashstate_function(tracking_address))
            logger.debug("Tracking id {} is state {}".format(str(tracking_id), str(state)))
            return hashstate == self.serializer.serialize(state)
        except ValueError as ve:
            raise NoTrackingException("No tracking exists with that id")

    def tracking_history(self, tracking_id):
        logger.info("Searching for tracking id {} history".format(str(tracking_id.hex())))
        entries = self.contract_interactor.search_logs('UnitMoved',
                                                       {'fromBlock': self.from_block, 'toBlock': self.to_block,
                                                        'argument_filters': {
                                                            'trackingId': tracking_id
                                                        }})
        res = []
        for e in entries:
            tx = self.gateway.get_transaction(e['transactionHash'])
            block = self.gateway.get_block(resolve_block_number(e['blockNumber']))
            res.append({'hash_state': e['args']['hashEstado'].hex(),
                        'message': e['args']['message'],
                        'block_number': resolve_block_number(e['blockNumber']),
                        'tx': e['transactionHash'].hex(),
                        'participant': tx['from'],
                        'datetime': datetime.fromtimestamp(block['timestamp']),
                        'timestamp': block['timestamp'],
                        })
        if len(res) == 0:
            raise NoTrackingException("No tracking exists with that id")
        return res

    def tracking_is_ended(self, tracking_id):
        logger.info("Evaluating if track is ended {}".format(str(tracking_id.hex())))
        entries = self.contract_interactor.search_logs('TrackingEnded',
                                                       {'fromBlock': self.from_block, 'toBlock': self.to_block,
                                                        'argument_filters': {
                                                            'trackingId': tracking_id
                                                        }
                                                        })
        return len(entries) > 0


class RoleChecker:
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, contract_interactor):
        self.contract_interactor = contract_interactor

    def is_role(self, role):
        logger.info("Checking if {} is role".format(str(role)))
        function = self.contract_interactor.get_function('isRole')
        res = self.contract_interactor.call_function(function(role.encode('utf-8')))
        logger.debug("Is role {}? {}".format(str(role), str(res)))
        return res


class TrackStarterNoWorkflow:
    contract_interactor = ...  # type: ContractInteractor
    serializer = DataSerializer()

    def __init__(self, contract_interactor, from_account):
        self.contract_interactor = contract_interactor
        self.owner = from_account

    def start(self, tracking_id, estado, next_role, message):
        serialized_estado = self.serializer.serialize(estado)
        function = self.contract_interactor.get_function('startTrack')
        logger.info("Starting track without workflow id: {} / state : {} / serialized state : {} /"
                    " next_role :  {} / message : {}"
                    .format(str(tracking_id), str(estado), str(serialized_estado), str(next_role), str(message)))
        res = self.contract_interactor.execute_function(
            function(tracking_id, serialized_estado, message, next_role.encode()), {'from': self.owner})
        return res


class TrackForwarderNoWorkflow:
    contract_interactor = ...  # type: ContractInteractor
    serializer = DataSerializer()

    def __init__(self, contract_interactor, owner, tracking_associator):
        self.tracking_associator = tracking_associator
        self.contract_interactor = contract_interactor
        self.owner = owner

    def move_forward(self, tracking_id, next_role, message, estado=None):
        if estado is None:
            estado = b''
        else:
            estado = self.serializer.serialize(estado)
        function = self.contract_interactor.get_function('forward')
        tracking_address = self.tracking_associator.tracking_contract(tracking_id)
        logger.info("Moving forward without workflow tracking id : {} / state : {} / next_role : {} / message : {}"
                    .format(str(tracking_id), str(estado), str(next_role), str(message)))
        res = self.contract_interactor.execute_function(function(tracking_address, estado, message, next_role.encode()),
                                                        {'from': self.owner})
        return res


class TrackingLister:
    contract_interactor = ...  # type: ContractInteractor
    gateway = ...  # type: Gateway
    from_block = 0
    to_block = 'latest'

    def __init__(self, **kwargs):

        for k, v in kwargs.items():
            setattr(self, k, v)

    def count(self):
        c = getattr(self, 'track_count', len(self.all()))
        self.track_count = c
        return self.track_count

    def all(self):
        logger.info("Searching for all tracks")
        entries = self.contract_interactor.search_logs('TrackingStarted',
                                                       {'fromBlock': self.from_block, 'toBlock': self.to_block, })
        res = []
        for e in entries:
            tx = self.gateway.get_transaction(e['transactionHash'])
            block = self.gateway.get_block(resolve_block_number(e['blockNumber']))
            res.append({
                'block_number': resolve_block_number(e['blockNumber']),
                'participant': tx['from'],
                'tracking_id': hex32bytes_string(e['args']['trackingId']),
                'tx': e['transactionHash'].hex(),
                'datetime': datetime.fromtimestamp(block['timestamp']),
                'timestamp': block['timestamp'],
            })
        self.track_count = len(res)
        res.sort(key=lambda k: k['block_number'])
        return res


class OwnerChecker:
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def is_owner(self, address):
        logger.info("Checking if {} is owner".format(str(address)))
        function = self.contract_interactor.get_function('isOwner')
        res = self.contract_interactor.call_function(function(address))
        logger.debug("Is owner {}? {}".format(str(address), str(res)))
        return res


class PendingStartsChecker:
    gateway = ...  # type: Gateway
    contract_interactor = ...  # type: ContractInteractor

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def check_pending(self, tracking_id):
        pending = self.gateway.pending_transactions()
        for p in pending:
            function, params = self.contract_interactor.decode_function_input(p)
            if params['trackingId'] == tracking_id:
                return True
        return False


class RoleAuthorizationChecker:
    contract_interactor = ...  # type: ContractInteractor
    tracking_association = ...  # type: TrackingAssociation

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def is_authorized_for_next_step(self, tracking_id, address):
        tracking_address = self.tracking_association.tracking_contract(tracking_id)
        logger.info("Checking if {} is authorized for next step for {}".format(str(address), str(tracking_id)))
        function = self.contract_interactor.get_function('participantIsAuthorizedForNextStep')
        res = self.contract_interactor.call_function(function(tracking_address, address))
        logger.debug("Is {} authorized? {}".format(str(address), str(res)))
        return res
