"""
Copyright 2019 de la Dirección General de Sistemas Informáticos – Secretaría Legal y Técnica - Nación.

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see http://www.gnu.org/licenses/
"""
import logging
from unittest import TestCase
import hashlib

# Create your tests here.
from web3.exceptions import ValidationError

from dapp.services import TrackStarter, ParticipantsLister, ParticipantAdding, WorkflowStepAdding, WorkflowLister, \
    TrackStatus, TrackForwarder, NoTrackingException, TrackingAssociation, TrackStarterNoWorkflow, \
    TrackForwarderNoWorkflow, TrackingLister, InvalidTrackingIdException, RolesLister, CompressSerializer, OwnerChecker, \
    PendingStartsChecker, RoleAuthorizationChecker
from gateway.gateway import Gateway, HTTPProviderGenerator, ContractInteractor, ContractDeployer, \
    InvalidTransactionException, HTTPFastLogSearcher
from dapp.test_settings import *

logger = logging.getLogger('tests-logger')


class ContractInteractorSimpleMock:

    def get_function(self, name):
        def some_function(a, b):
            pass

        return some_function

    def execute_function(self, function, params, error_msg=""):
        return "0x948394834"


class StartTrackTest(TestCase):
    def setUp(self):
        prov = HTTPProviderGenerator(NODE_URL)
        self.gateway = Gateway(prov, IS_POA)
        self.owner = self.gateway.get_node_accounts()[0]
        deploy = ContractDeployer(self.gateway)
        receipt = deploy.deploy_contract(ABI, BYTECODE, deploy.estimate_deploy(ABI, BYTECODE), self.owner, True)
        mock = ContractInteractorSimpleMock()
        self.track_starter = TrackStarter(mock, self.owner)
        self.tracking_id = hashlib.sha256("tracking_id".encode('utf-8')).digest()
        self.estado = {'attr': 'sarlanga'}

    def test_start_track_mock(self):
        tx = self.track_starter.start(self.tracking_id, self.estado)
        self.assertTrue(is_hex(tx))

    def test_start_track_invalid(self):
        tracking_id_invalid = "DFJLSDFJSDFJLSDKFJSDFJSLDFJSDLFJSDLFKJSDKLFJSDKLFSDJFLSDJFLKSDJFKLSJLSFJLL".encode()
        self.assertRaises(InvalidTrackingIdException, self.track_starter.start, tracking_id=tracking_id_invalid,
                          state={})


class GatewayMockListerEmpty:
    def get_contract(self, a, b):
        return None

    def get_events(self, contract):
        return None


class GatewayMockList:
    def all(self):
        return [
            {'account': '0xca35b7d915458ef540ade6068dfe2f44e8fa733c', 'roles': ['START']},
            {'account': '0x14723a09acff6d2a60dcdf7aa4aff308fddc160c', 'roles': ['NEXT_STEP']},
            {'account': '0x4b0897b0513fdc7c541b6d9d7e929c4e5364d2db', 'roles': ['MORE_STEP']},
            {'account': '0x583031d1113ad414f02576bd6afabfb302140225', 'roles': ['ALMOST']},
            {'account': '0xdd870fa1b7c4700f2bd7f44238821c26f7392148', 'roles': ['DONE']},
        ]


def is_hex(s):
    try:
        int(s, 16)
        return True
    except ValueError:
        return False


class ParticipantAddingTest(TestCase):

    def setUp(self) -> None:
        prov = HTTPProviderGenerator(NODE_URL)
        self.gateway = Gateway(prov, IS_POA)
        self.owner = self.gateway.get_node_accounts()[0]
        deploy = ContractDeployer(self.gateway)
        receipt = deploy.deploy_contract(ABI, BYTECODE, deploy.estimate_deploy(ABI, BYTECODE), self.owner, True)
        contract_interactor = ContractInteractor(self.gateway, ABI, receipt['contractAddress'])
        self.participant_adding_service = ParticipantAdding(contract_interactor, self.owner)
        self.participant_lister = ParticipantsLister(contract_interactor, receipt['blockNumber'])

    def test_add_participant(self):
        # no me importa si es el propio owner
        account = self.gateway.get_node_accounts()[0]
        otro_account = self.gateway.get_node_accounts()[1]
        res = self.participant_adding_service.add_participant(account, 'START_TEST')
        receipt = self.gateway.get_receipt(res, True)
        self.assertTrue(is_hex(res))
        self.assertEqual(receipt['transactionHash'].hex(), res)
        self.assertTrue(self.participant_lister.is_participant(account))
        self.assertFalse(self.participant_lister.is_participant(otro_account))
        all_part = self.participant_lister.all()
        self.assertEqual(len(all_part), 1)

    def test_add_invalid_role(self):
        account = self.gateway.get_node_accounts()[0]
        # roles must be 32 bytes top, else is invalid
        role = 'START_TESTFGHJDFGDFGDFGJKDFGHDFJGKHDFGJKDFGHKDFKGHDFJKDFGDJFGDFGHDHFGDFGHDFJKGDFJKGHDFGJKJKDHFG'
        self.assertRaises(ValidationError, self.participant_adding_service.add_participant, address=account, role=role)


class WorkflowStepTest(TestCase):

    def setUp(self) -> None:
        prov = HTTPProviderGenerator(NODE_URL)
        self.gateway = Gateway(prov, IS_POA)
        self.owner = self.gateway.get_node_accounts()[0]
        deploy = ContractDeployer(self.gateway)
        receipt = deploy.deploy_contract(ABI, BYTECODE, deploy.estimate_deploy(ABI, BYTECODE), self.owner, True)
        contract_interactor = ContractInteractor(self.gateway, ABI, receipt['contractAddress'])
        self.workflowstep_adder = WorkflowStepAdding(contract_interactor, self.owner)
        self.workflow_lister = WorkflowLister(contract_interactor)
        self.participant_adding_service = ParticipantAdding(contract_interactor, self.owner)

    def test_add_workflow_step_no_participants(self):
        self.assertRaises(InvalidTransactionException, self.workflowstep_adder.add_workflow_step,
                          message='Arranca pa', role='START')
        workflow_list = self.workflow_lister.all()
        self.assertTrue(len(workflow_list) == 0)

    def test_add_workflow_step_one_participant(self):
        account = self.gateway.get_node_accounts()[0]
        self.gateway.get_receipt(self.participant_adding_service.add_participant(account, 'STARTER'), True)
        self.assertTrue(len(self.workflow_lister.all()) == 0)
        res = self.workflowstep_adder.add_workflow_step('Arranca pa', 'STARTER')
        rcpt = self.gateway.get_receipt(res, True)
        self.assertTrue(len(self.workflow_lister.all()) == 1)


class TrackTestHelpers:
    def set_up(self):
        prov = HTTPProviderGenerator(NODE_URL)
        self.gateway = Gateway(prov, IS_POA)
        self.owner = self.gateway.get_node_accounts()[0]
        logger.info('Deploying contract with account {}'.format(str(self.owner)))
        deploy = ContractDeployer(self.gateway)
        receipt = deploy.deploy_contract(ABI, BYTECODE, deploy.estimate_deploy(ABI, BYTECODE), self.owner, True)
        logger.info('Contract deployed address: {}'.format(str(receipt['contractAddress'])))
        self.contract_interactor = ContractInteractor(self.gateway, ABI, receipt['contractAddress'],
                                                      HTTPFastLogSearcher(gateway=self.gateway, url=NODE_URL))
        self.workflowstep_adder = WorkflowStepAdding(self.contract_interactor, self.owner)
        self.workflow_lister = WorkflowLister(self.contract_interactor, receipt['blockNumber'])
        self.participant_adding_service = ParticipantAdding(self.contract_interactor, self.owner)
        self.track_status = TrackStatus(self.contract_interactor,
                                        TrackingAssociation(self.contract_interactor))

    def add_participant(self):
        account = self.gateway.get_node_accounts()[1]
        res = [self.participant_adding_service.add_participant(account, "STARTER"),
               self.participant_adding_service.add_participant(account, "FST_NEXTER"),
               self.participant_adding_service.add_participant(account, "SND_NEXTER"),
               self.participant_adding_service.add_participant(account, "ENDER")]

        # wait for the txs to be sealed
        for r in res:
            self.gateway.get_receipt(r, True)

    def add_workflow_steps(self):
        res = [self.workflowstep_adder.add_workflow_step("Starting track", "STARTER"),
               self.workflowstep_adder.add_workflow_step("Moved from A", "FST_NEXTER"),
               self.workflowstep_adder.add_workflow_step("Almost ending", "SND_NEXTER"),
               self.workflowstep_adder.add_workflow_step("Ending track", "ENDER")]

        # wait for txs to be sealed
        for r in res:
            self.gateway.get_receipt(r, True)


class SimpleFullTrackTest(TestCase, TrackTestHelpers):
    # in this test, i make a simple workflow case, 1 participant, 1 role, 4 workflow steps and a track which goes
    # through them

    def setUp(self) -> None:
        self.set_up()

    def test_simple_valid_workflow(self):
        self.add_participant()
        self.add_workflow_steps()
        tracking_id = hashlib.sha256("test_track_start".encode('utf-8')).digest()
        estado = {"attr1": "value", "attr2": "value"}
        account = self.gateway.get_node_accounts()[1]

        # arranco trackeo
        logger.info("Starting track account {}".format(str(account)))
        track_starter = TrackStarter(self.contract_interactor, account)
        res = track_starter.start(tracking_id, estado)
        logger.info("Tx hash start track {}".format(str(res)))
        self.assertTrue(is_hex(res))
        rpt = self.gateway.get_receipt(res, True)
        block = self.gateway.get_block(rpt['blockNumber'])
        logger.info("Checking last message")
        mensaje = self.track_status.last_message(tracking_id)
        logger.info("Last message {}".format(str(mensaje)))
        bool_state = self.track_status.is_this_state(tracking_id, estado)
        self.assertTrue(bool_state)
        self.assertEqual("Starting track", mensaje)
        logger.info("Looking tracking history")
        track_history = self.track_status.tracking_history(tracking_id)
        logger.info("Track history {}".format(str(track_history)))
        self.assertTrue(len(track_history) == 1)
        self.assertEqual(track_history[0]['tx'], res)
        self.assertEqual(track_history[0]['message'], mensaje)
        self.assertEqual(track_history[0]['hash_state'], self.track_status.serializer.serialize(estado).hex())
        self.assertEqual(track_history[0]['participant'], account)
        self.assertEqual(track_history[0]['block_number'], rpt['blockNumber'])
        self.assertEqual(track_history[0]['timestamp'], block['timestamp'])
        self.assertFalse(self.track_status.tracking_is_ended(tracking_id))

        # primer movimiento
        track_forwarder = TrackForwarder(self.contract_interactor, account,
                                         TrackingAssociation(self.contract_interactor))
        estado = {"attr1": "value", "attr2": "value2"}
        res = track_forwarder.move_forward(tracking_id, estado)
        rpt = self.gateway.get_receipt(res, True)
        block = self.gateway.get_block(rpt['blockNumber'])
        logger.info("Checking last message")
        mensaje = self.track_status.last_message(tracking_id)
        logger.info("Last message {}".format(str(mensaje)))
        self.assertEqual(mensaje, "Moved from A")
        bool_state = self.track_status.is_this_state(tracking_id, estado)
        logger.info("Looking tracking history")
        track_history = self.track_status.tracking_history(tracking_id)
        logger.info("Track history {}".format(str(track_history)))
        self.assertEqual(len(track_history), 2)
        self.assertTrue(bool_state)
        self.assertEqual(track_history[1]['tx'], res)
        self.assertEqual(track_history[1]['message'], mensaje)
        self.assertEqual(track_history[1]['participant'], account)
        self.assertEqual(track_history[1]['block_number'], rpt['blockNumber'])
        self.assertEqual(track_history[1]['timestamp'], block['timestamp'])
        self.assertFalse(self.track_status.tracking_is_ended(tracking_id))

        # segundo movimiento
        res = track_forwarder.move_forward(tracking_id)
        rpt = self.gateway.get_receipt(res, True)
        block = self.gateway.get_block(rpt['blockNumber'])
        logger.info("Checking last message")
        mensaje = self.track_status.last_message(tracking_id)
        logger.info("Last message {}".format(str(mensaje)))
        self.assertEqual(mensaje, "Almost ending")
        bool_state = self.track_status.is_this_state(tracking_id, estado)
        logger.info("Looking tracking history")
        track_history = self.track_status.tracking_history(tracking_id)
        logger.info("Track history {}".format(str(track_history)))
        self.assertEqual(len(track_history), 3)
        self.assertTrue(bool_state)
        self.assertEqual(track_history[2]['tx'], res)
        self.assertEqual(track_history[2]['message'], mensaje)
        self.assertEqual(track_history[2]['participant'], account)
        self.assertEqual(track_history[2]['block_number'], rpt['blockNumber'])
        self.assertEqual(track_history[2]['timestamp'], block['timestamp'])
        self.assertFalse(self.track_status.tracking_is_ended(tracking_id))

        # tercer movimiento
        res = track_forwarder.move_forward(tracking_id)
        rpt = self.gateway.get_receipt(res, True)
        block = self.gateway.get_block(rpt['blockNumber'])
        self.assertRaises(NoTrackingException, self.track_status.is_this_state, tracking_id=tracking_id, state=estado)
        logger.info("Looking tracking history")
        track_history = self.track_status.tracking_history(tracking_id)
        logger.info("Track history {}".format(str(track_history)))
        self.assertEqual(len(track_history), 4)
        self.assertEqual(track_history[3]['tx'], res)
        self.assertEqual(track_history[3]['message'], self.track_status.last_message(tracking_id))
        self.assertEqual(track_history[3]['participant'], account)
        self.assertEqual(track_history[3]['block_number'], rpt['blockNumber'])
        self.assertEqual(track_history[3]['timestamp'], block['timestamp'])
        self.assertTrue(self.track_status.tracking_is_ended(tracking_id))


class SimpleFullNoWorkflowTrackTest(TestCase, TrackTestHelpers):

    def setUp(self) -> None:
        self.set_up()

    def test_simple_valid_workflow(self):
        self.add_participant()
        tracking_id = hashlib.sha256("test_track_start".encode('utf-8')).digest()
        estado = {"attr1": "value", "attr2": "value"}
        account = self.gateway.get_node_accounts()[1]

        # arranco trackeo
        track_starter = TrackStarterNoWorkflow(self.contract_interactor, account)
        res = track_starter.start(tracking_id, estado, "FST_NEXTER", "Starting track")
        self.assertTrue(is_hex(res))
        self.gateway.get_receipt(res, True)
        mensaje = self.track_status.last_message(tracking_id)
        bool_state = self.track_status.is_this_state(tracking_id, estado)
        self.assertTrue(bool_state)
        self.assertEqual("Starting track", mensaje)
        track_history = self.track_status.tracking_history(tracking_id)
        self.assertTrue(len(track_history) == 1)
        self.assertEqual(track_history[0]['tx'], res)
        self.assertEqual(track_history[0]['message'], mensaje)
        self.assertEqual(track_history[0]['hash_state'], self.track_status.serializer.serialize(estado).hex())

        # primer movimiento
        track_forwarder = TrackForwarderNoWorkflow(self.contract_interactor, account,
                                                   TrackingAssociation(self.contract_interactor))
        res = track_forwarder.move_forward(tracking_id, "SND_NEXTER", "Moved from A")
        self.gateway.get_receipt(res, True)
        mensaje = self.track_status.last_message(tracking_id)
        self.assertEqual(mensaje, "Moved from A")
        bool_state = self.track_status.is_this_state(tracking_id, estado)
        track_history = self.track_status.tracking_history(tracking_id)
        self.assertEqual(len(track_history), 2)
        self.assertTrue(bool_state)
        self.assertEqual(track_history[1]['tx'], res)
        self.assertEqual(track_history[1]['message'], mensaje)

        # segundo movimiento
        res = track_forwarder.move_forward(tracking_id, "ENDER", "Almost ending")
        self.gateway.get_receipt(res, True)
        mensaje = self.track_status.last_message(tracking_id)
        self.assertEqual(mensaje, "Almost ending")
        bool_state = self.track_status.is_this_state(tracking_id, estado)
        track_history = self.track_status.tracking_history(tracking_id)
        self.assertEqual(len(track_history), 3)
        self.assertTrue(bool_state)
        self.assertEqual(track_history[2]['tx'], res)
        self.assertEqual(track_history[2]['message'], mensaje)

        # tercer movimiento
        res = track_forwarder.move_forward(tracking_id, "", "the end")
        self.gateway.get_receipt(res, True)
        self.assertRaises(NoTrackingException, self.track_status.is_this_state, tracking_id=tracking_id, state=estado)
        track_history = self.track_status.tracking_history(tracking_id)
        self.assertEqual(len(track_history), 4)
        self.assertEqual(track_history[3]['tx'], res)
        self.assertEqual(track_history[3]['message'], self.track_status.last_message(tracking_id))
        self.assertTrue(self.track_status.tracking_is_ended(tracking_id))


class TrackingListerTest(TestCase, TrackTestHelpers):
    def setUp(self) -> None:
        self.set_up()

    def test_simple_valid_workflow(self):
        self.add_participant()
        tracking_ids = [hashlib.sha256("idtrack1".encode('utf-8')).digest(),
                        hashlib.sha256("idtrack2".encode('utf-8')).digest(),
                        hashlib.sha256("idtrack3".encode('utf-8')).digest(),
                        hashlib.sha256("idtrack4".encode('utf-8')).digest()]
        estado = {"attr1": "value", "attr2": "value"}
        account = self.gateway.get_node_accounts()[1]

        # arranco trackeo
        track_starter = TrackStarterNoWorkflow(self.contract_interactor, account)
        res = [track_starter.start(tracking_ids[0], estado, "FST_NEXTER", "Starting track"),
               track_starter.start(tracking_ids[1], estado, "FST_NEXTER", "Starting track"),
               track_starter.start(tracking_ids[2], estado, "FST_NEXTER", "Starting track"),
               track_starter.start(tracking_ids[3], estado, "FST_NEXTER", "Starting track")]

        self.gateway.get_receipt(res[3], True)

        lister = TrackingLister(contract_interactor=self.contract_interactor, gateway=self.gateway)
        trackings = lister.all()
        self.assertEqual(4, lister.count())
        for t in trackings:
            self.assertIn('participant', t)
            self.assertIn('tracking_id', t)
            self.assertIn('block_number', t)
            self.assertIn('tx', t)
            self.assertIn('datetime', t)
            self.assertIn('timestamp', t)


class RolesListerTest(TestCase, TrackTestHelpers):
    def setUp(self) -> None:
        self.set_up()

    def test_list_roles(self):
        accounts = self.gateway.get_node_accounts()[:4]
        accounts2 = self.gateway.get_node_accounts()[5:9]
        res = []
        for a in accounts:
            res.append(self.participant_adding_service.add_participant(a, "SOME_ROLE"))

        for a in accounts2:
            res.append(self.participant_adding_service.add_participant(a, "ANOTHER_ROLE"))

        # wait for the txs to be sealed
        for r in res:
            self.gateway.get_receipt(r, True)

        role_lister = RolesLister(self.contract_interactor)
        roles = role_lister.all()
        for k, v in roles.items():
            self.assertGreater(len(v), 0)


class TestCompressSerializer(TestCase):

    def setUp(self) -> None:
        self.compressor = CompressSerializer()

    def test_serialize(self):
        data = {'asdasdasd': 'asdsadasdsa', 'attr1': "val1", 'attr2': "valu2", 'attr3': "val3",
                "fsdffjsdfdjldjflsdklf": "dsfjksdhjkfsdhfjksdhfsdhfdhfsdhfksdhfsdjkfhdjks",
                "sarlanga": "fjkgdfjkghfjkhdfjkgfhgkdhgkdfhgkjdfghdfjkghdfjkghdfjkghdfjkghdfkghdfjk"}
        comp = self.compressor.serialize(data)
        self.assertDictEqual(data, self.compressor.deserialize(comp))


class TestOwnerChecker(TestCase, TrackTestHelpers):
    def setUp(self) -> None:
        self.set_up()

    def test_is_owner(self):
        checker = OwnerChecker(contract_interactor=self.contract_interactor)
        owner = self.gateway.get_node_accounts()[0]
        res = checker.is_owner(owner)
        self.assertTrue(res)

    def test_is_not_owner(self):
        checker = OwnerChecker(contract_interactor=self.contract_interactor)
        not_owner = self.gateway.get_node_accounts()[1]
        res = checker.is_owner(not_owner)
        self.assertFalse(res)


class PendingStartTrackTest(TestCase, TrackTestHelpers):
    def setUp(self) -> None:
        self.set_up()

    def test_start_track_pendings(self):
        account = self.gateway.get_node_accounts()[0]
        self.gateway.get_receipt(self.participant_adding_service.add_participant(account, "STARTER"), True)
        self.add_workflow_steps()
        tracking_id = hashlib.sha256("test_track_start".encode('utf-8')).digest()
        estado = {"attr1": "value", "attr2": "value"}
        account = self.gateway.get_node_accounts()[1]
        self.gateway.unlock_account(account, 'Clave123!')

        logger.info("Starting track account {}".format(str(account)))
        track_starter = TrackStarter(self.contract_interactor, account)
        pending_checker = PendingStartsChecker(gateway=self.gateway)
        are_pendings = pending_checker.check_pending(tracking_id)
        self.assertFalse(are_pendings)
        res = track_starter.start(tracking_id, estado)
        are_pendings = pending_checker.check_pending(tracking_id)
        self.assertTrue(are_pendings)
        rpt = self.gateway.get_receipt(res, True)
        pending_checker.check_pending(tracking_id)
        self.assertFalse(are_pendings)


class WorkflowIsDefinedTest(TestCase, TrackTestHelpers):
    def setUp(self) -> None:
        self.set_up()

    def test_workflow_is_defined(self):
        workflow_lister = WorkflowLister(self.contract_interactor, 0, 'latest')
        self.add_participant()
        self.add_workflow_steps()
        self.assertTrue(workflow_lister.workflow_is_defined())

    def test_workflow_is_not_defined(self):
        workflow_lister = WorkflowLister(self.contract_interactor, 0, 'latest')
        self.assertFalse(workflow_lister.workflow_is_defined())


class AccountForwardAuthorization(TestCase, TrackTestHelpers):
    def setUp(self) -> None:
        self.set_up()

    def test_account_is_authorized(self):
        self.add_participant()
        self.add_workflow_steps()
        tracking_id = hashlib.sha256("test_track_start".encode('utf-8')).digest()
        estado = {"attr1": "value", "attr2": "value"}
        account = self.gateway.get_node_accounts()[1]
        unauthorized_account = self.gateway.get_node_accounts()[2]
        authorization_checker = RoleAuthorizationChecker(contract_interactor=self.contract_interactor,
                                                         tracking_association=TrackingAssociation(self.contract_interactor))

        # arranco trackeo
        logger.info("Starting track account {}".format(str(account)))
        track_starter = TrackStarter(self.contract_interactor, account)
        res = track_starter.start(tracking_id, estado)
        rpt = self.gateway.get_receipt(res, True)
        is_authorized = authorization_checker.is_authorized_for_next_step(tracking_id, account)
        is_not_authorized = authorization_checker.is_authorized_for_next_step(tracking_id, unauthorized_account)
        self.assertTrue(is_authorized)
        self.assertFalse(is_not_authorized)
